#!/usr/bin/env python
# Copyright (c) 2013 Cloudera, Inc. All rights reserved.
"""
Functionality for manipulating Linux cgroups.
"""

import logging
import os
import re
import subprocess
import tempfile

LOG = logging.getLogger(__name__)
PROC_CGROUPS = "/proc/cgroups"
PROC_MOUNTS = "/proc/mounts"

# TODO: Need to model better relationships between subsystems, resources,
# and resource conversions (see Process.evaluate_resources()).

# Maps subsystems to cgroups pseudofiles.
SUBSYSTEMS = {"blkio"  : ["blkio.weight"],
              "cpu"    : ["cpu.shares"],
              "memory" : ["memory.limit_in_bytes"]}

def init_capabilities():
  """
  Returns an initialized cgroups capabilities dictionary. Should match the
  layout of the Avro heartbeat.
  """
  caps = dict()
  caps["has_blkio"] = False
  caps["has_cpu"] = False
  caps["has_memory"] = False
  caps["default_blkio_weight"] = -1
  caps["default_cpu_shares"] = -1
  caps["default_memory_limit_in_bytes"] = -1
  caps["writable_cgroup_dot_procs"] = False
  return caps

class CGroupManager(object):
  """
  Encapsulates cgroups management functionality.
  """

  def __init__(self, path):
    self._path = path
    self._mounted_subsystems = []

    # Probe capabilities
    self._empty_capabilities = init_capabilities()
    self._capable_subsystems = []
    self.capabilities = self._probe()
    if self.capabilities:
      for cap, value in self.capabilities.items():
        subsys = self._avro_cap_to_subsystem(cap)
        if subsys and value:
          self._capable_subsystems.append(subsys)

    # Make subsystem directories.
    for subsys in self._capable_subsystems:
      subsys_path = os.path.join(path, subsys)
      if os.path.isdir(subsys_path):
        LOG.info("Reusing %s" % (subsys_path,))
      else:
        LOG.info("Created %s" % (subsys_path,))
        os.makedirs(subsys_path)

  def _do_mount(self, subsys, path):
    """
    Mount the desired cgroups subsystem at the desired path.

    Returns True if successful, False otherwise.
    """
    LOG.debug("Mounting cgroups subsystem %s to %s" % (subsys, path))
    popen = subprocess.Popen(["mount", "-t", "cgroup", "-o", subsys,
                              "cm_cgroups", path], stderr=subprocess.PIPE)
    _, stderr = popen.communicate()
    if popen.returncode:
      LOG.info("Failed to mount cgroups subsystem %s to %s, rc: %s stderr: %s"
               % (subsys, path, popen.returncode, stderr))
    return popen.returncode == 0

  def _do_umount(self, path):
    """
    Unmount the cgroups subsystem found at the desired path.

    Returns True if successful, False otherwise.
    """
    LOG.debug("Unmounting cgroups subsystem from %s" % (path,))
    popen = subprocess.Popen(["umount", path], stderr=subprocess.PIPE)
    _, stderr = popen.communicate()
    if popen.returncode:
      LOG.info("Failed to unmount cgroups subsystem from %s, rc: %s stderr: %s"
               % (path, popen.returncode, stderr))
    return popen.returncode == 0

  def _subsystem_to_avro_cap(self, subsystem):
    """
    Returns the converted result of a subsystem name into an avro capability,
    or None if the subsystem isn't represented by a capability.
    """
    assert subsystem in SUBSYSTEMS
    cap_key = "has_" + subsystem
    if cap_key in self._empty_capabilities:
      return cap_key
    return None

  def _avro_cap_to_subsystem(self, avro_cap):
    """
    Returns the converted result of an avro capability into a subsystem name,
    or None if the capability does not represent a subsystem.
    """
    assert avro_cap in self._empty_capabilities
    subsys = re.sub("^has_", "", avro_cap)
    if subsys in SUBSYSTEMS:
      return subsys
    return None

  def _probe_mount(self, subsystem, probe_writable_cgroup_dot_procs,
                   pseudofiles):
    """
    Mounts a subsystem in a temporary location and probes selected
    pseudofiles for capabilities.

    If requested, will also check whether cgroup.procs is writable.
    """
    caps = dict()
    d = tempfile.mkdtemp()
    try:
      if not self._do_mount(subsystem, d):
        return
      try:
        cap_key = self._subsystem_to_avro_cap(subsystem)
        caps[cap_key] = True

        # Probe the pseudofiles.
        for pf in pseudofiles:
          # Should match the Avro heartbeat.
          cap_key = "default_" + pf.replace(".", "_")

          pf_path = os.path.join(d, pf)
          LOG.debug("Opening cgroups pseudofile %s" % (pf_path))
          opened_pf = open(pf_path)
          try:
            caps[cap_key] = int(opened_pf.read())
          finally:
            opened_pf.close()
        if probe_writable_cgroup_dot_procs:
          cg_dot_procs_path = os.path.join(d, "cgroup.procs")
          LOG.debug("Testing cgroup.procs %s" % (cg_dot_procs_path))
          caps["writable_cgroup_dot_procs"] = os.access(cg_dot_procs_path,
                                                        os.W_OK)
      finally:
        self._do_umount(d)
    finally:
      os.rmdir(d)

    return caps

  def _get_proc_mounts_matchers(self):
    """
    Return a dictionary of matcher->subsystem suitable for matching against
    the contents of /proc/mounts.
    """
    subsys_matchers = dict()
    for subsys in self._capable_subsystems:
      subsys_path = os.path.join(self._path, subsys)
      # Example: "none /path/to/mount fstype..."
      matcher = re.compile("\S+\s+(%s)\s+cgroup" % (subsys_path,))
      subsys_matchers[matcher] = subsys
    return subsys_matchers

  def _probe(self):
    """
    Probe the system for cgroups capabilities.

    Temporarily mounts one subsystem at a time to fully probe capabilities
    (note: we can mount them all together in one mount, but that fails if
    cgroups are already mounted somewhere on the system.

    Returns a dictionary with the results, or None if no support is found.
    """

    # Look for /proc/cgroups
    if not os.path.exists(PROC_CGROUPS):
      LOG.debug("Cannot find %s; no cgroups capabilities found" %
                (PROC_CGROUPS,))
      return None

    # We need to see whether cgroup.procs is writable. We can do this on
    # any mounted subsystem, so we'll try on the first one we find.
    probe_writable_cgroup_dot_procs = True

    # Prepare the /proc/cgroups subsystem regex matchers.
    subsys_matchers = dict()
    for subsys in SUBSYSTEMS:
      matcher = re.compile("%s\s+" % (subsys,))
      subsys_matchers[matcher] = subsys

    caps = init_capabilities()
    f = open(PROC_CGROUPS, "r")
    try:
      # Look for relevant subsystems in /proc/cgroups. When we find one,
      # mount it temporarily to extract other capabilities.
      for line in f:
        more_caps = None
        for matcher, subsys in subsys_matchers.items():
          if matcher.match(line):
            LOG.debug("Found cgroups subsystem: %s" % (subsys,))
            more_caps = self._probe_mount(subsys,
                                          probe_writable_cgroup_dot_procs,
                                          SUBSYSTEMS[subsys])
            probe_writable_cgroup_dot_procs = False
            if more_caps:
              caps = dict(caps.items() + more_caps.items())
    finally:
      f.close()

    return caps

  def enable(self):
    """
    Mount all supported cgroups subsystems, reusing any that are already
    mounted.
    """

    # Do we need to mount anything?
    subsys_needing_mount = (set(self._capable_subsystems) -
                            set(self._mounted_subsystems))
    if len(subsys_needing_mount) == 0:
      return

    LOG.debug("Need to mount %s" % (subsys_needing_mount,))

    # Find out if our subsystems are already mounted.
    subsys_matchers = self._get_proc_mounts_matchers()
    f = open(PROC_MOUNTS)
    try:
      for line in f:
        for matcher, subsys in subsys_matchers.items():
          m = matcher.match(line)
          if m and subsys in subsys_needing_mount:
            LOG.info("Reusing mount %s" % (m.group(1),))
            subsys_needing_mount.remove(subsys)
            self._mounted_subsystems.append(subsys)
    finally:
      f.close()

    # Mount whatever is left.
    for subsys in subsys_needing_mount:
      subsys_path = os.path.join(self._path, subsys)
      LOG.info("Mounting subsystem %s at %s" % (subsys, subsys_path))
      if not self._do_mount(subsys, subsys_path):
        continue
      self._mounted_subsystems.append(subsys)

  def disable(self):
    """
    Unmount all mounted subsystems.
    """

    # Do we need to unmount anything?
    if len(self._mounted_subsystems) == 0:
      return

    LOG.debug("Need to unmount %s" % (self._mounted_subsystems,))

    # We think we need to unmount something, but let's use /proc/mounts as
    # our source of truth, in case someone else unmounted something.
    subsys_needing_unmount = set()
    subsys_matchers = self._get_proc_mounts_matchers()
    f = open(PROC_MOUNTS)
    try:
      for line in f:
        for matcher, subsys in subsys_matchers.items():
          m = matcher.match(line)
          if m and subsys in self._mounted_subsystems:
            LOG.debug("Found %s mounted" % (m.group(1),))
            subsys_needing_unmount.add(subsys)
    finally:
      f.close()

    # Remove the mount from our list even if unmount fails.
    for subsys in subsys_needing_unmount:
      subsys_path = os.path.join(self._path, subsys)
      LOG.info("Unmounting subsystem %s at %s" % (subsys, subsys_path))
      self._do_umount(subsys_path)
      self._mounted_subsystems.remove(subsys)

  def create_group(self, name):
    """
    Creates a new set of cgroups (one per subsystem) using the given name.

    Returns True if at least one group was created, False otherwise.
    """
    retval = False
    if len(self._mounted_subsystems) == 0:
      return retval

    for group_path in self.get_groups(name):
      if not os.path.isdir(group_path):
        LOG.info("Creating cgroup %s" % (group_path,))
        try:
          os.mkdir(group_path)
          retval = True
        except:
          LOG.exception("Failed to make group %s" % (name,))
      else:
        LOG.info("Reusing cgroup %s" % (group_path,))
    return retval

  def destroy_group(self, name):
    """
    Destroys a set of cgroups (one per subsystem) with the given name.
    """
    if len(self._mounted_subsystems) == 0:
      return

    for group_path in self.get_groups(name):
      if os.path.isdir(group_path):
        LOG.info("Destroying cgroup %s" % (group_path,))
        try:
          os.rmdir(group_path)
        except:
          LOG.exception("Failed to remove group %s" % (name,))
      else:
        LOG.info("The cgroup %s is already destroyed" % (group_path,))

  def configure_group(self, name, subsys, resource):
    """
    Configures a cgroup with the given name using the given subsystem and
    resource.

    TODO: Dynamic resources will incur a pseudofile open() + read() with
    every heartbeat. This can be avoided by caching the current value and
    rereading it at startup.

    Returns True if the group was reconfigured, False otherwise.
    """

    assert subsys in SUBSYSTEMS

    if (len(self._mounted_subsystems) == 0 or
        subsys not in self._mounted_subsystems):
      return False

    # XXX: Assumes there's only one pseudofile per subsystem.
    pf = SUBSYSTEMS[subsys][0]
    pf_path = os.path.join(self._path, subsys, name, pf)
    f = open(pf_path, "r+")
    try:
      # Only update if the old value doesn't match the new.
      oldval = int(f.read().strip())
      if (oldval == resource):
        LOG.debug("Values for cgroup pseudofile %s match (%d)" %
                  (pf_path, resource))
      else:
        LOG.info("Reconfiguring cgroup pseudofile %s with value %d" %
                 (pf_path, resource))
        f.write(str(resource))
    finally:
      f.close()
    return True

  def get_groups(self, name):
    """
    Returns a list of cgroups that are applicable for the given name.
    """
    groups = []
    for subsys in self._mounted_subsystems:
      groups.append(os.path.join(self._path, subsys, name))
    return groups
