#!/bin/bash

# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.

# Time marker for both stderr and stdout
date; date 1>&2

cloudera_config=`dirname $0`
cloudera_config=`cd "$cloudera_config/../common"; pwd`
. ${cloudera_config}/cloudera-config.sh

# attempt to find java
locate_java_home

# load the parcel environment
source_parcel_environment

if [ "4" = "$CDH_VERSION" ]; then
  . ${cloudera_config}/cdh4-default
  HDFS_BIN=$HADOOP_HDFS_HOME/bin/hdfs

  # Disable IPv6. This is automatically done by bin/hadoop for CDH3.
  export HADOOP_OPTS="-Djava.net.preferIPv4Stack=true $HADOOP_OPTS"
elif [ "3" = "$CDH_VERSION" ] || [ "-3" = "$CDH_VERSION" ]; then
  HDFS_BIN=$CDH_HADOOP_HOME/bin/hadoop
  export HADOOP_HOME=$CDH_HADOOP_HOME
  echo "using $HADOOP_HOME as HADOOP_HOME"
else
  echo "ERROR: Unsupported version $CDH_VERSION"
  exit 1
fi

echo "using $JAVA_HOME as JAVA_HOME"
echo "using $CDH_VERSION as CDH_VERSION"
echo "using $CONF_DIR as CONF_DIR"
echo "using $SECURE_USER as SECURE_USER"
echo "using $SECURE_GROUP as SECURE_GROUP"

set_hadoop_classpath

# debug
set -x

# local replacement of substitution variables.
# i.e. master doesn't know final resting place of configuration files
#   which may require absolute path (e.g. dfs_allows)
perl -pi -e "s#{{CMF_CONF_DIR}}#$CONF_DIR#g" $CONF_DIR/*.xml

if [ -e $CONF_DIR/topology.py ]; then
  perl -pi -e "s#{{CMF_CONF_DIR}}#$CONF_DIR#g" $CONF_DIR/topology.py
fi

if [ -e $CONF_DIR/log4j.properties ]; then
  perl -pi -e "s#{{CMF_CONF_DIR}}#$CONF_DIR#g" $CONF_DIR/log4j.properties
fi

# Make sure topology.py and cloudera_manager_agent_fencer.py (where applicable) are executable
OUTPUT=`find $CONF_DIR -maxdepth 1 -name "*.py"`
if [ "$OUTPUT" != "" ]; then
  chmod +x $CONF_DIR/*.py
fi

# if HADOOP_IDENT_STRING is not set, $USER will be used instead. There have
# been situations where $USER is not set for su which results in
# hadoop.id.str ending up to be an empty string so we're setting it here
# explicitly
export HADOOP_IDENT_STRING="hdfs"

# Secure datanode requires additonal env vars.
if [ -n "$SECURE_DATANODE" ]; then
  if [ "4" = "$CDH_VERSION" ]; then
    export HADOOP_SECURE_DN_USER=$SECURE_USER
    echo "using $HADOOP_SECURE_DN_USER as HADOOP_SECURE_DN_USER"
    # We have decided to put jsvc in different locations on Centos and Sles.
    # So we need to check what we set by default and at the alternative location
    # too to make sure we have the right one.
    if [[ ! -e $JSVC_HOME/jsvc ]]; then
      export JSVC_HOME=/usr/lib/bigtop-utils
      if [[ ! -e $JSVC_HOME/jsvc ]]; then
        echo "No jsvc file found."
        exit 1
      fi
    fi
    echo "using $JSVC_HOME as JSVC_HOME"
  elif [ "3" = "$CDH_VERSION" ] || [ "-3" = "$CDH_VERSION" ]; then
    export HADOOP_DATANODE_USER=$SECURE_USER
    echo "using $HADOOP_DATANODE_USER as HADOOP_DATANODE_USER"
  fi
  # We're running as a daemon. The agent spawns us as root:root but HDFS_BIN
  # will impersonate another user, so file ownership must be reset. 
  chown -R $SECURE_USER:$SECURE_GROUP $CONF_DIR
fi

acquire_kerberos_tgt hdfs.keytab

if [ "validate-writable-empty-dirs" = "$1" ]; then
  if [ -z "$VALIDATE_DIRS" ]; then
    echo "No namenode dirs specified." # unexpected state; CMF bug.
    exit 1
  fi
  for validate_dir in $VALIDATE_DIRS; do
    if [ -e $validate_dir ]; then
      if [[ ! -d $validate_dir || ! -w $validate_dir ]]; then
        echo "Validation failed: $validate_dir must be a directory and writable by hdfs."
        exit 1
      fi
      NON_EMPTY_DIR=`find -L "$validate_dir" -maxdepth 0 -not -empty`
      if [ -n "$NON_EMPTY_DIR" ]; then
        echo "Validation failed: $validate_dir exists, but is not empty"
        exit 1
      fi
    else
      # resource was not made
      echo "Validation failed: $validate_dir does not exist and cannot be created."
      exit 1
    fi
  done
  echo "Validation of $VALIDATE_DIRS succeeded."
  exit 0
elif [ "file-operation" = "$1" ]; then
  TARGET_DIR=$2
  FILENAME=$3
  COMMAND=$4
  FULL_PATH=$TARGET_DIR/$FILENAME
  if [ "WRITE" = "$COMMAND" ]; then
    cat > $FULL_PATH <<< $FILENAME
    RET=$?
    if [ $RET -eq 0 ]; then
      echo "Successfully wrote $FULL_PATH"
    else
      echo "Failed to write $FULL_PATH"
    fi
 elif [ "CONFIRM_EXISTS" = "$COMMAND" ]; then
   # Confirm existence of a file by trying to write the same file
   # when clobbering is not allowed. Writing is preferred over reading
   # for NFS, where sync on write should be configured.
   set -o noclobber
   cat > $FULL_PATH <<< $FILENAME
   RET=$?
   if [ $RET -eq 1 ]; then
     # Return code of 1 when cannot overwrite existing file
     echo "Successfully confirmed that file already exists: $FULL_PATH"
     RET=0
   elif [ $RET -eq 0 ]; then
     echo "File did not exist: $FULL_PATH. Make sure that $TARGET_DIR is shared and is configured to sync on write."
     rm $FULL_PATH
     RET=1
   else
     echo "Failed to confirm whether file exists: $FULL_PATH"
   fi
   set +o noclobber
 elif [ "DELETE" = "$COMMAND" ]; then
    rm $FULL_PATH
    RET=$?
    if [ $RET -eq 0 ]; then
      echo "Successfully deleted $FULL_PATH"
    else
      echo "Failed to delete $FULL_PATH"
    fi
  else
    echo "Illegal command: $COMMAND"
    RET=1
  fi
  exit $RET
elif [ "bootstrap" = "$1" ]; then
  exec $HDFS_BIN --config $CONF_DIR namenode -bootstrapStandby -nonInteractive
elif [ "failover" = "$1" ]; then
  ACTIVE="Failover failed: Can't failover to an active service"
  NS=$2
  FROM_NN=$3
  TO_NN=$4
  FORCE=$5
  if [ "$FORCE" = "true" ]; then
    FORCE_FAILOVER="--forceactive"
  fi
  OUTPUT=$( $HDFS_BIN --config $CONF_DIR haadmin -ns $NS -failover $FORCE_FAILOVER $FROM_NN $TO_NN 2>&1 )
  RET=$?
  if [ $RET -eq 0 ]; then
    echo "Failed over from $FROM_NN to $TO_NN of nameservice $NS"
    exit 0
  elif ( grep -q "$ACTIVE" <<< "$OUTPUT" ); then
    echo "Failover from $FROM_NN to $TO_NN of nameservice $NS unnecessary as $TO_NN is already active."
    exit 2
  else
    echo "Unable to failover from $FROM_NN to $TO_NN of nameservice $NS; see stderr log."
    if [ $RET -eq 2 ]; then
      # Return code of 2 is already used to indicate an acceptable failure.
      exit 1
    else
      exit $RET
    fi
  fi
elif [ "transition-to-active" = $1 ]; then
  if [ -z "$2" || -z "$3" ]; then
    echo "Nameservice and NameNode ID must be specified." # unexpected state; CM bug.
    exit 1
  fi
  NS=$2
  NN=$3
  exec $HDFS_BIN --config $CONF_DIR haadmin -ns $NS -transitionToActive $NN
elif [ "initializeSharedEdits" = "$1" ]; then
  exec $HDFS_BIN --config $CONF_DIR namenode -initializeSharedEdits -nonInteractive
elif [ "initialize-znode" = "$1" ]; then
  exec $HDFS_BIN --config $CONF_DIR zkfc -formatZK -nonInteractive
elif [ "format-namenode" = "$1" ]; then
  # hadoop namenode -format helpfully double-checks your
  # request if the storage directories already exist.
  # In CM, they often do exist but are empty (because
  # we've pre-created them).
  # So, we force the format, as long as all the directories
  # are empty.
  if [ -z "$DFS_STORAGE_DIRS" ]; then
    echo "No storage dirs specified." # unexpected state; CM bug.
    exit 1
  fi
  for dfsdir in $DFS_STORAGE_DIRS; do
    if [ -e $dfsdir ]; then
      if [ ! -d $dfsdir ]; then
        echo "Refusing to format.  $dfsdir is not a directory."
        exit 1
      fi
      NON_EMPTY_DIR=`find -L "$dfsdir" -maxdepth 0 -not -empty`
      if [ -n "$NON_EMPTY_DIR" ]; then
        echo "Refusing to format: $dfsdir exists and is non-empty"
	cm_error_and_exit HDFS_DFS_DIR_NOT_EMPTY
      fi
    fi
  done

  CLUSTER_ARGS=""
  if [ $# -eq 2 ]; then
    CLUSTER_ARGS="-clusterId $2"
  fi

  # XXX: Can we use -nonInteractive now?
  # XXX: The use of exec in a pipeline has no effect, but should make it clear
  #      that when this hack is removed, the exec should remain.
  /usr/bin/yes Y 2> /dev/null | exec $HDFS_BIN --config $CONF_DIR namenode -format $CLUSTER_ARGS
elif [ "monitor-decommission" = "$1" ]; then
  # $2 is the list of names to look for in a '|' delimited list.
  # $3 is the exact NameNode to monitor, in the format of hdfs://host:port
  # Extract hostnames of DNs being decommissioned
  WAIT=0
  while [ $WAIT -eq 0 ]; do
    # There's nothing special about 15 seconds. It's just attempting to
    # strike a balance between too much polling and avoiding a large
    # artificial delay in the case that the operation is quick.
    sleep 15
    # Yay for robust interfaces.
    # We used to check for exact host names till CM 4.0.
    # But after https://jira.cloudera.com/browse/OPSAPS-9202, we just wait till any host is being decommissioned.
    $HDFS_BIN --config $CONF_DIR dfsadmin -fs $3 -report | grep -q 'Decommission in progress'
    WAIT=$?
  done
elif [ "jnSyncWait" = "$1" ]; then
  # Wait for all the journal nodes to get in sync with the edits of this namenode.
  # $2 is the NN's /jmx URL which is parsed to find out how many JNs are in sync.
  # $3 is the number of JN's in HDFS. We wait till this number of JNs are in sync.
  SYNC_COUNT=0
  ADD_IMPORTS="import urllib2; from urllib2_kerberos import HTTPKerberosAuthHandler;"
  while [ "$SYNC_COUNT" != "$3" ]; do
    sleep 15
    SYNC_COUNT=`python -c "$ADD_IMPORTS print urllib2.build_opener(HTTPKerberosAuthHandler).open('$2').read()" | grep IsOutOfSync | grep -c false`
  done
elif [ "nnRpcWait" = "$1" ]; then
  # Wait for namenode to start responding to RPCs, by checking that we can
  # open the URL specified in $2.
  WAIT=1
  ADD_IMPORTS="import urllib2; from urllib2_kerberos import HTTPKerberosAuthHandler;"
  while [ $WAIT -ne 0 ]; do
    sleep 15
    python -c "$ADD_IMPORTS urllib2.build_opener(HTTPKerberosAuthHandler).open('$2')"
    WAIT=$?
  done
elif [ "monitor-upgrade" = "$1" ]; then
  RET=0
  WAIT=0
  while [ $WAIT -eq 0]; do
    # See above
    sleep 15
    REPORT=`$HDFS_BIN --config $CONF_DIR dfsadmin -upgradeProgress status`
    # XXX: Check $? here and do something?
    echo $REPORT | grep -q -E "Upgrade for version .* has been completed"
    WAIT=$?
    RET=$WAIT
  done
  exit $RET
elif [ "finalize-upgrade" = "$1" ]; then
  exec $HDFS_BIN --config $CONF_DIR dfsadmin -finalizeUpgrade
elif [ "mkdir" = "$1" ]; then
  if [ $# -ne 5 ]; then
    echo "mkdir not called with expected number of arguments. (4)"
    exit 1
  fi

  DIR=$2
  USER=$3
  GROUP=$4
  PERMS=$5

  RET=0
  $HDFS_BIN --config $CONF_DIR dfs -test -d "$DIR"
  if [ $? -eq 0 ]; then
    echo "Directory $DIR already exists."
  else
    echo "Directory $DIR does not currently exist."

    # Starting with CDH 4.1, HDFS won't automatically create the parent dir
    # if -p is not specified.
    PARENT_ARG=""
    if [ "3" != "$CDH_VERSION" ] && [ "-3" != "$CDH_VERSION" ]; then
      PARENT_ARG="-p"
    fi
    $HDFS_BIN --config $CONF_DIR dfs -mkdir $PARENT_ARG "$DIR"
    if [ $? -eq 0 ]; then
      echo "Created directory $DIR."
    else
      echo "Unable to create directory $DIR; see stderr log."
      exit 1
    fi

    $HDFS_BIN --config $CONF_DIR dfs -chown "$USER":"$GROUP" "$DIR"
    if [ $? -eq 0 ]; then
      echo "Changed ownership of $DIR to $USER:$GROUP."
    else
      echo "Unable to change ownership of $DIR to $USER:$GROUP; see stderr log."
      RET=1
    fi
    $HDFS_BIN --config $CONF_DIR dfs -chmod "$PERMS" "$DIR"
    if [ $? -eq 0 ]; then
      echo "Changed permissions of $DIR to $PERMS."
    else
      echo "Unable to change permissions of $DIR to $PERMS; see stderr log."
      RET=1
    fi
  fi

  exit $RET
elif [ "namenode" = "$1" -o \
       "secondarynamenode" = "$1" -o \
       "datanode" = "$1" ]; then
  # Set hadoop security and audit log appenders. These are set here instead
  # of being hardcoded in the log4j template because we only want the hadoop
  # daemons to use them.
  HADOOP_OPTS="-Dsecurity.audit.logger=$HADOOP_SECURITY_LOGGER $HADOOP_OPTS"
  export HADOOP_OPTS="-Dhdfs.audit.logger=$HADOOP_AUDIT_LOGGER $HADOOP_OPTS"

  exec $HDFS_BIN --config $CONF_DIR "$@"
else
  exec $HDFS_BIN --config $CONF_DIR "$@"
fi
