// (c) Copyright 2011-2012 Cloudera, Inc. All rights reserved.
define([
  'knockout',
  'cloudera/EventSearch'
], function(ko, EventSearch) {

  var FilterModel = EventSearch.FilterModel,
    FilterViewModel = EventSearch.FilterViewModel;

  describe('FilterModel', function() {

    var fakeFilterViewModel, filterModel;

    beforeEach(function() {
      fakeFilterViewModel = {
        filters: jasmine.createSpy()
      };
      filterModel = new FilterModel(fakeFilterViewModel);
    });
    
    it('defines some attributes', function() {
      expect(filterModel.filter).toBeDefined();
      expect(filterModel.compareType).toBeDefined();
      expect(filterModel.value).toBeDefined();
      expect(filterModel.values).toBeDefined();
      // Computed values.
      expect(filterModel.filterName).toBeDefined();
      expect(filterModel.compareTypeName).toBeDefined();
    });

    it('is complete when all values are satisfied for single value case', function() {
      filterModel.filter('filter');
      expect(filterModel.complete()).toBeFalsy();
      filterModel.compareType('==');
      expect(filterModel.complete()).toBeFalsy();
      filterModel.value('value');
      expect(filterModel.complete()).toBeTruthy();
    });

    it('is complete when all values are satisfied for multi-value case', function() {
      filterModel.filter('filter');
      expect(filterModel.complete()).toBeFalsy();
      filterModel.compareType('==');
      expect(filterModel.complete()).toBeFalsy();
      filterModel.values(['value1', 'value2']);
      expect(filterModel.complete()).toBeTruthy();
    });

    it('will only be activated if not already active and called by button', function() {
      // Create a button element for use in this method.
      var $fakeButton = $('<div class="btn"></div>').appendTo($('body'));
      var $fakeNotbutton = $('<div></div>').appendTo($('body'));
      var eventData = {
        target: $fakeNotbutton[0]
      };

      // Use this to track if anything actually fired.
      var changed = false;
      filterModel.active.subscribe(function() {
        changed = true;
      });

      expect(filterModel.active()).toBeFalsy();
      filterModel.activate(filterModel, eventData);
      expect(filterModel.active()).toBeTruthy();
      expect(changed).toBeTruthy();

      // Now with something that *is* a button.
      filterModel.active(false);
      changed = false;
      eventData.target = $fakeButton[0];
      filterModel.activate(filterModel, eventData);
      expect(filterModel.active()).toBeFalsy();
      expect(changed).toBeFalsy();

      // Now do it when it is already active.
      eventData.target = $fakeNotbutton[0];
      filterModel.active(true);
      changed = false;
      filterModel.activate(filterModel, eventData);
      expect(filterModel.active()).toBeTruthy();
      expect(changed).toBeFalsy();

      // Clean up after ourselves.
      $fakeButton.remove();
      $fakeNotbutton.remove();
    });

    it('knows how to compare against others of its type in single value case', function() {
      var otherFilterModel = new FilterModel();
      otherFilterModel.filter('filter');
      otherFilterModel.compareType('==');
      otherFilterModel.value('value');
      filterModel.filter('filter');
      filterModel.compareType('==');
      filterModel.value('value');

      expect(filterModel.compareTo(otherFilterModel)).toBeTruthy();
      otherFilterModel.filter('bad');
      expect(filterModel.compareTo(otherFilterModel)).toBeFalsy();
      otherFilterModel.filter(filterModel.filter());
      otherFilterModel.compareType('bad');
      expect(filterModel.compareTo(otherFilterModel)).toBeFalsy();
      otherFilterModel.compareType(filterModel.compareType());
      otherFilterModel.value('bad');
      expect(filterModel.compareTo(otherFilterModel)).toBeFalsy();
      otherFilterModel.value(filterModel.value());
      expect(filterModel.compareTo(otherFilterModel)).toBeTruthy();
    });

    it('knows how to compare against others of its type in multi-value case with different order', function() {
      var otherFilterModel = new FilterModel();
      otherFilterModel.filter('filter');
      otherFilterModel.compareType('==');
      otherFilterModel.values(['value1', 'value2']);
      filterModel.filter('filter');
      filterModel.compareType('==');
      // Note that the above was ['value1', 'value2'].
      filterModel.values(['value2', 'value1']);
      expect(filterModel.compareTo(otherFilterModel)).toBeTruthy();
    });

    it('can compare to falsy correctly', function() {
      expect(filterModel.compareTo(null)).toBeFalsy();
      expect(filterModel.compareTo(undefined)).toBeFalsy();
    });

    it('will not notify observers of change to values property if array does not change', function() {
      var changed = jasmine.createSpy();
      filterModel.values.subscribe(changed);
      filterModel.values([]);
      expect(changed).wasNotCalled();
      filterModel.values([1]);
      expect(changed).wasCalled();
      changed.reset();
      filterModel.values([2]);
      expect(changed).wasCalled();
      changed.reset();
      filterModel.values([]);
      expect(changed).wasCalled();
    });

    it('sets filter properly by name', function() {
      var fakeFilter1 = { propertyName: 'catpants' };
      var fakeFilter2 = { propertyName: 'doggiehat' };
      fakeFilterViewModel.filters.andReturn([fakeFilter1, fakeFilter2]);
      filterModel.filterName('doggiehat');
      expect(filterModel.filter()).toEqual(fakeFilter2);
      filterModel.filterName('horsepoo');
      expect(filterModel.filter()).toEqual(null);
    });

    it('tracks filter name as a property', function() {
      filterModel.filter({ propertyName: 'catpants' });
      expect(filterModel.filterName()).toEqual('catpants');
      filterModel.filter({ propertyName: 'doggiehat' });
      expect(filterModel.filterName()).toEqual('doggiehat');
    });

    it('is careful about setting compareType by name when filter is null', function() {
      filterModel.filter(null);
      // If this line doesn't blow up, this test is successful.
      filterModel.compareTypeName('catpants');
      expect(filterModel.compareTypeName()).toBeFalsy();
    });

    it('sets compareType by name', function() {
      var fakeCompareType1 = { name: 'catpants' };
      var fakeCompareType2 = { name: 'doggiehat' };
      filterModel.filter({
        compareTypes: [fakeCompareType1, fakeCompareType2]
      });
      filterModel.compareTypeName('doggiehat');
      expect(filterModel.compareType()).toEqual(fakeCompareType2);
      filterModel.compareTypeName('horsepoo');
      expect(filterModel.compareType()).toEqual(null);
    });

    it('provides values as comma-delimited string', function() {
      expect(filterModel.value()).toEqual('');
      filterModel.values(['catpants', 'doggiehat']);
      expect(filterModel.value()).toEqual('catpants,doggiehat');
    });

    it('allows values to be set as comma-delimited string', function() {
      expect(filterModel.values().length).toEqual(0);
      filterModel.value('catpants,doggiehat');
      expect(filterModel.values().length).toEqual(2);
      var values = filterModel.values();
      expect(values[0]).toEqual('catpants');
      expect(values[1]).toEqual('doggiehat');
    });
  });

  describe('FilterViewModel', function() {

    var fakeEventSearch, filterViewModel;

    beforeEach(function() {
      fakeEventSearch = {
        activeFilter: ko.observable(),
        appliedFilters: ko.observableArray(),
        appendFilter: jasmine.createSpy('appendFilter'),
        loading: jasmine.createSpy('loading'),
        filters: jasmine.createSpy('filters'),
        removeFilter: jasmine.createSpy('removeFilter'),
        findFilter: jasmine.createSpy('findFilter'),
        compareTypeFromName: function() {}
      };
      filterViewModel = new FilterViewModel(fakeEventSearch);
    });

    it('defines some attributes', function() {
      expect(filterViewModel.activeFilter).toBeDefined();
      expect(filterViewModel.appliedFilters).toBeDefined();
      expect(filterViewModel.loading).toBeDefined();
      expect(filterViewModel.filters).toBeDefined();
      expect(filterViewModel.removeFilter).toBeDefined();
      expect(filterViewModel.findFilter).toBeDefined();
    });

    it('appends filters', function() {
      spyOn(fakeEventSearch, 'compareTypeFromName').andReturn('compareType');
      expect(filterViewModel.appliedFilters().length).toEqual(0);
      fakeEventSearch.appendFilter.andReturn(true);
      var result = filterViewModel.appendFilter('filter', 'compareTypeName', 'value');
      expect(result).toBeTruthy();
      expect(fakeEventSearch.appendFilter).wasCalled();
    });

    it('converts single value to singleton array for values', function() {
      spyOn(fakeEventSearch, 'compareTypeFromName').andReturn('compareType');
      filterViewModel.appendFilter('filter', 'compareTypeName', 'catpants');
      expect(fakeEventSearch.appendFilter).wasCalled();
      var filterModel = fakeEventSearch.appendFilter.mostRecentCall.args[0];
      expect(filterModel.values().length).toEqual(1);
      expect(filterModel.values()[0]).toEqual('catpants');
    });

    it('will convert appended filter compare type if given string', function() {
      spyOn(fakeEventSearch, 'compareTypeFromName').andReturn('compareType');
      fakeEventSearch.appendFilter.andReturn(true);
      var result = filterViewModel.appendFilter('filter', 'compareTypeName', 'value');
      expect(result).toBeTruthy();
      expect(fakeEventSearch.compareTypeFromName).wasCalled();

      fakeEventSearch.compareTypeFromName.reset();
      result = filterViewModel.appendFilter('filter', {}, 'value');
      expect(result).toBeTruthy();
      expect(fakeEventSearch.compareTypeFromName).wasNotCalled();
    });

    it('will fail if compare type is not found', function() {
      spyOn(fakeEventSearch, 'compareTypeFromName').andReturn(null);
      var result = filterViewModel.appendFilter('filter', 'compareTypeName', 'value');
      expect(result).toBeFalsy();
      expect(fakeEventSearch.compareTypeFromName).wasCalled();
    });

    it('can deactivate filters', function() {
      spyOn(filterViewModel, 'activeFilter').andReturn(null);
      filterViewModel.deactivateFilters();
      expect(filterViewModel.activeFilter).wasCalled();

      filterViewModel.activeFilter.reset();
      var fakeFilter = {
        active: jasmine.createSpy('active')
      };
      filterViewModel.activeFilter.andReturn(fakeFilter);
      filterViewModel.deactivateFilters();
      expect(filterViewModel.activeFilter).wasCalled();
      expect(fakeFilter.active).wasCalledWith(false);
    });

    it('can add another filter while one is active', function() {
      spyOn(filterViewModel, 'deactivateFilters');
      spyOn(filterViewModel, 'appendFilter');
      filterViewModel.addAnother();
      expect(filterViewModel.deactivateFilters).wasCalled();
      expect(filterViewModel.appendFilter).wasCalled();
    });
  });

  describe('EventSearch', function() {

    var $eventFilters, $eventList, staticFilters, dataSourceSpy,
      exportEventsSpy, eventSearch, goodResponse;

    beforeEach(function() {
      jasmine.Ajax.useMock();
      // Dummy success JSON to return from mock responses.
      goodResponse = {
        status: 200,
        responseText: JSON.stringify({success: true})
      };
      // Generate the mock data that we're sending in to the event
      // search instance.
      $eventFilters = $('<div></div>').appendTo($('body'));
      $eventList = $('<div></div>').appendTo($('body'));
      staticFilters = [];
      dataSourceSpy = jasmine.createSpy();
      exportEventsSpy = jasmine.createSpy();

      eventSearch = new EventSearch($eventFilters[0], $eventList[0],
        'test/filters/url', [], dataSourceSpy, exportEventsSpy);
    });

    afterEach(function() {
      $eventFilters.remove();
      $eventList.remove();
    });

    it('defines some attributes', function() {
      expect(eventSearch.error).toBeDefined();
      expect(eventSearch.filterContainer).toBeDefined();
      expect(eventSearch.filters).toBeDefined();
      expect(eventSearch.staticFilters).toBeDefined();
      expect(eventSearch.appliedFilters).toBeDefined();
      expect(eventSearch.filterPath).toBeDefined();
      expect(eventSearch.activeFilter).toBeDefined();
      expect(eventSearch.eventList).toBeDefined();
      expect(eventSearch.events).toBeDefined();
      expect(eventSearch.loading).toBeDefined();
      expect(eventSearch.dataSource).toBeDefined();
      expect(eventSearch.exportEvents).toBeDefined();
      expect(eventSearch.viewModel).toBeDefined();
    });

    it('applies bindings upon instantiation', function() {
      spyOn(ko, 'applyBindings');
      eventSearch = new EventSearch($eventFilters[0], $eventList[0],
        'test/filters/url', [], dataSourceSpy, exportEventsSpy);
      expect(ko.applyBindings.callCount).toEqual(2);
      var args = ko.applyBindings.argsForCall[0];
      expect(args[0]).toEqual(eventSearch.viewModel);
      expect(args[1]).toEqual($eventFilters[0]);
      args = ko.applyBindings.argsForCall[1];
      expect(args[1]).toEqual($eventList[0]);
    });

    it('exposes a subset of itself as a view model to the event list', function() {
      spyOn(ko, 'applyBindings');
      eventSearch = new EventSearch($eventFilters[0], $eventList[0],
        'test/filters/url', [], dataSourceSpy, exportEventsSpy);
      expect(ko.applyBindings.callCount).toEqual(2);
      var args = ko.applyBindings.argsForCall[1];
      var viewModel = args[0];
      expect(viewModel.loading).toEqual(eventSearch.loading);
      expect(viewModel.error).toEqual(eventSearch.error);
      expect(viewModel.events).toEqual(eventSearch.events);
    });

    it('exposes a hasFilterNamed function to the event list', function() {
      spyOn(ko, 'applyBindings');
      eventSearch = new EventSearch($eventFilters[0], $eventList[0],
        'test/filters/url', [], dataSourceSpy, exportEventsSpy);
      expect(ko.applyBindings.callCount).toEqual(2);
      var args = ko.applyBindings.argsForCall[1];
      var viewModel = args[0];
      eventSearch.filters([{
        propertyName: 'catpants'
      }, {
        propertyName: 'doggiehat'
      }]);
      expect(viewModel.hasFilterNamed('catpants')).toBeTruthy();
      expect(viewModel.hasFilterNamed('horsepoo')).toBeFalsy();
    });

    it('can get a compare type from its name', function() {
      var compareType1 = { name: 'compareType1' };
      var compareType2 = { name: 'compareType2' };
      var filter = {
        compareTypes: [compareType1, compareType2]
      };
      var compareType = eventSearch.compareTypeFromName(filter, 'compareType2');
      expect(compareType).toEqual(compareType2);
      compareType = eventSearch.compareTypeFromName(filter, 'unknown');
      expect(compareType).toBeFalsy();
    });

    it('getHead and getTail work when there are no events', function() {
      expect(eventSearch.length()).toEqual(0);
      expect(eventSearch.getHead()).toBeNull();
      expect(eventSearch.getTail()).toBeNull();
    });

    it('getHead and getTail work when there are events', function() {
      eventSearch.events([1, 2, 3]);
      expect(eventSearch.length()).toEqual(3);
      expect(eventSearch.getHead()).toEqual(1);
      expect(eventSearch.getTail()).toEqual(3);
    });

    it('can splice events to the end', function() {
      eventSearch.events([1, 2, 3]);
      expect(eventSearch.length()).toEqual(3);
      eventSearch.splice(0);
      expect(eventSearch.length()).toEqual(0);
      eventSearch.events([1, 2, 3]);
      eventSearch.splice(3, 0, 4);
      expect(eventSearch.length()).toEqual(4);
      var events = eventSearch.events();
      expect(events[0]).toEqual(1);
      expect(events[1]).toEqual(2);
      expect(events[2]).toEqual(3);
      expect(events[3]).toEqual(4);
    });

    it('correctly appends filters', function() {
      var filter = new FilterModel();
      expect(eventSearch.appliedFilters().length).toEqual(0);
      eventSearch.appendFilter(filter);
      expect(eventSearch.appliedFilters().length).toEqual(1);
    });

    it('does not append if any filters are duplicates', function() {
      var filter = new FilterModel();
      filter.filter('filter');
      filter.compareType('compareType');
      filter.value('value');
      expect(eventSearch.appliedFilters().length).toEqual(0);
      eventSearch.appendFilter(filter);
      expect(eventSearch.appliedFilters().length).toEqual(1);
      // Try to add exact same filter.
      var filter2 = new FilterModel();
      filter2.filter('filter');
      filter2.compareType('compareType');
      filter2.value('value');
      eventSearch.appendFilter(filter2);
      expect(eventSearch.appliedFilters().length).toEqual(1);
    });

    it('sets filter to active if not complete', function() {
      var filter = new FilterModel();
      expect(filter.complete()).toBeFalsy();
      expect(filter.active()).toBeFalsy();
      eventSearch.appendFilter(filter);
      expect(filter.active()).toBeTruthy();
    });

    it('tracks which filter is the active one and allows only one', function() {
      var filter1 = new FilterModel();
      filter1.filter('filter1');
      filter1.compareType('compareType1');
      filter1.value('value1');
      var filter2 = new FilterModel();
      filter2.filter('filter2');
      filter2.compareType('compareType2');
      filter2.value('value2');
      eventSearch.appendFilter(filter1);
      eventSearch.appendFilter(filter2);
      expect(eventSearch.appliedFilters().length).toEqual(2);
      filter1.active(true);
      expect(filter1.active()).toBeTruthy();
      expect(filter2.active()).toBeFalsy();
      filter2.active(true);
      expect(filter1.active()).toBeFalsy();
      expect(filter2.active()).toBeTruthy();
    });

    it('removes incomplete active filters when deactivated', function() {
      var filter1 = new FilterModel();
      filter1.filter('filter1');
      eventSearch.appendFilter(filter1);
      expect(filter1.active()).toBeTruthy();
      expect(eventSearch.activeFilter()).toEqual(filter1);
      expect(eventSearch.appliedFilters().length).toEqual(1);
      eventSearch.deactivateAllFilters();
      expect(eventSearch.appliedFilters().length).toEqual(0);
      expect(eventSearch.activeFilter()).toBeFalsy();
    });

    it('tracks filters and removes uncompleted deactivated ones', function() {
      spyOn(eventSearch, 'removeFilter');
      var filter1 = new FilterModel();
      eventSearch.appendFilter(filter1);
      expect(filter1.active()).toBeTruthy();
      expect(filter1.complete()).toBeFalsy();
      expect(eventSearch.appliedFilters().length).toEqual(1);
      filter1.active(false);
      expect(eventSearch.removeFilter).wasCalled();
    });

    it('fires an event when an active, completed filter is set inactive', function() {
      spyOn($, 'publish');
      var filter1 = new FilterModel();
      eventSearch.appendFilter(filter1);
      expect($.publish).wasNotCalled();
      expect(filter1.active()).toBeTruthy();
      expect(filter1.complete()).toBeFalsy();
      expect(eventSearch.appliedFilters().length).toEqual(1);
      filter1.filter('filter1');
      filter1.compareType('compareType1');
      filter1.value('value1');
      expect(filter1.complete()).toBeTruthy();
      filter1.active(false);
      expect($.publish).wasCalledWith('eventsearch.appliedfilters.changed');
    });

    it('knows how to remove a filter', function() {
      var filter1 = new FilterModel();
      filter1.filter('filter1');
      filter1.compareType('compareType1');
      filter1.value('value1');
      var filter2 = new FilterModel();
      filter2.filter('filter2');
      filter2.compareType('compareType2');
      filter2.value('value2');
      eventSearch.appendFilter(filter1);
      eventSearch.appendFilter(filter2);
      expect(eventSearch.appliedFilters().length).toEqual(2);

      // Start tracking how many times the array is changed.
      var count = 0;
      eventSearch.appliedFilters.subscribe(function() {
        count += 1;
      });
      eventSearch.removeFilter(filter1);
      expect(eventSearch.appliedFilters().length).toEqual(1);
      expect(count).toEqual(1);
      expect(filter1.active()).toBeFalsy();
      expect(eventSearch.activeFilter()).toEqual(null);
    });

    it('deactivates an active filter when it is removed', function() {
      var filter = new FilterModel();
      filter.filter('filter1');
      filter.compareType('compareType1');
      filter.value('value1');
      eventSearch.appendFilter(filter);
      filter.active(true);
      expect(eventSearch.activeFilter()).toEqual(filter);
      eventSearch.removeFilter(filter);
      expect(eventSearch.activeFilter()).toEqual(null);
    });

    it('correctly deactivates filters', function() {
      var filter = new FilterModel();
      eventSearch.appendFilter(filter);
      expect(filter.active()).toBeTruthy();
      eventSearch.deactivateAllFilters();
      expect(filter.active()).toBeFalsy();
    });

    it('correctly prepends events', function() {
      // Test the single event case.
      var event1 = new FilterModel();
      event1.name = 'event1';
      var result = eventSearch.prepend(event1);
      expect(result).toEqual(1);

      // Test the array case.
      var event2 = new FilterModel();
      event2.name = 'event2';
      var event3 = new FilterModel();
      event3.name = 'event3';
      result = eventSearch.prepend([event2, event3]);
      expect(result).toEqual(3);
      var events = eventSearch.events();
      expect(events[0].name).toEqual('event2');
      expect(events[1].name).toEqual('event3');
      expect(events[2].name).toEqual('event1');

      // Test the prepend to empty events case.
      eventSearch.events([]);
      expect(eventSearch.length()).toEqual(0);
      result = eventSearch.prepend([event1, event2, event3]);
      expect(result).toEqual(eventSearch.length());
    });

    it('correctly appends events', function() {
      // Test the single event case.
      var event1 = new FilterModel();
      event1.name = 'event1';
      var result = eventSearch.append(event1);
      expect(result).toEqual(1);

      // Test the array case.
      var event2 = new FilterModel();
      event2.name = 'event2';
      var event3 = new FilterModel();
      event3.name = 'event3';
      result = eventSearch.append([event2, event3]);
      expect(result).toEqual(3);
      var events = eventSearch.events();
      expect(events[0].name).toEqual('event1');
      expect(events[1].name).toEqual('event2');
      expect(events[2].name).toEqual('event3');

      // Test the append to empty events case.
      eventSearch.events([]);
      expect(eventSearch.length()).toEqual(0);
      result = eventSearch.append([event1, event2, event3]);
      expect(result).toEqual(eventSearch.length());
    });

    it('loads filters via a getJSON call', function() {
      spyOn(eventSearch, 'loadFiltersHandler');
      eventSearch.loadFilters();
      var request = mostRecentAjaxRequest();
      request.response(goodResponse);
      expect(eventSearch.loadFiltersHandler).wasCalled();
    });

    it('knows how to handle the filters response from the server', function() {
      eventSearch.filters([1, 2]);
      var responseFromServer = {
        message: 'OK',
        data: [4, 5, 6]
      };
      eventSearch.loadFiltersHandler(responseFromServer);
      expect(eventSearch.filters().length).toEqual(3);
      var filters = eventSearch.filters();
      expect(filters[0]).toEqual(4);
      expect(filters[1]).toEqual(5);
      expect(filters[2]).toEqual(6);
    });

    it('calls a callback when filters are loaded', function() {
      eventSearch.filters([1, 2]);
      var callback = jasmine.createSpy('load filters callback');
      eventSearch.loadFilters(callback);
      var request = mostRecentAjaxRequest();
      request.response(goodResponse);
      expect(callback).wasCalled();
    });

    it('has method for retrying the dataSource after a time', function() {
      var clear = true;
      var handler = jasmine.createSpy();
      var timeoutInMs = 5;
      runs(function() {
        eventSearch.retryHandlerAfterTime(clear, handler, timeoutInMs);
      });
      waitsFor(function() {
        return dataSourceSpy.callCount > 0;
      }, 'dataSource never called', timeoutInMs * 2);
      runs(function() {
        // Verify that dataSource was called with correct args.
        var args = dataSourceSpy.mostRecentCall.args;
        expect(args[0]).toEqual(eventSearch);
        expect(args[1]).toEqual(clear);
        expect(args[2]).toEqual(handler);
      });
    });

    it('knows how to handle errors when loading data', function() {
      // We're not providing a callback here so we don't have to
      // worry about async retries.
      var handler = eventSearch.makeLoadDataHandler(true);
      expect(eventSearch.error()).toBeFalsy();
      eventSearch.loading(true);

      handler(null);
      expect(eventSearch.error()).toBeFalsy();
      expect(eventSearch.loading()).toBeFalsy();

      eventSearch.loading(true);
      var errorMessage = 'catpants';
      handler(errorMessage);
      expect(eventSearch.error()).toEqual(errorMessage);
      expect(eventSearch.loading()).toBeFalsy();
    });

    it('knows how to call callbacks when loading data', function() {
      var callback = jasmine.createSpy().andReturn(false);
      spyOn(eventSearch, 'retryHandlerAfterTime');
      var handler = eventSearch.makeLoadDataHandler(true, callback);

      handler(null);
      expect(callback).wasCalled();
      expect(eventSearch.retryHandlerAfterTime).wasNotCalled();

      callback.andReturn(true);
      handler('error');
      expect(callback).wasCalledWith('error');
      expect(eventSearch.retryHandlerAfterTime).wasCalled();
    });

    it('knows how to load data', function() {
      var handler = 'handler';
      var callback = function() {};
      spyOn(eventSearch, 'makeLoadDataHandler').andReturn(handler);
      eventSearch.error('catpants error');
      expect(eventSearch.loading()).toBeFalsy();

      eventSearch.loadData(true, callback);
      expect(eventSearch.error()).toBeFalsy();
      expect(eventSearch.loading()).toBeTruthy();
      expect(dataSourceSpy).wasCalledWith(eventSearch, true, handler);

      eventSearch.error('catpants error');
      eventSearch.loading(false);
      dataSourceSpy.reset();
      eventSearch.loadData(callback);
      expect(eventSearch.error()).toBeFalsy();
      expect(eventSearch.loading()).toBeTruthy();
      expect(dataSourceSpy).wasCalledWith(eventSearch, undefined, handler);
    });

    it('can return all the applied filters', function() {
      var makeTestFilterModel = function(index) {
        var model = new FilterModel();
        model.filter({ propertyName: 'propertyName ' + index});
        model.compareType({ name: 'compareType ' + index });
        model.value('value ' + index);
        return model;
      };
      var model1 = makeTestFilterModel(1);
      var model2 = makeTestFilterModel(2);
      var model3 = makeTestFilterModel(3);
      // Create a special one that has 'values' set instead of 'value'.
      var model4 = new FilterModel();
      model4.filter({ propertyName: 'propertyName 4'});
      model4.compareType({ name: 'compareType 4' });
      model4.values(['value 4.1', 'value 4.2']);

      // The integrity check in getAppliedFilters should discard
      // model3 because of its null value.
      model3.value(null);
      // Add a static filter, it should come first. We don't care about
      // the rest of the properties for the purposes of this test.
      var staticModel = {
        values: ['value 0']
      };
      eventSearch.staticFilters = [staticModel];
      eventSearch.appliedFilters([model1, model2, model3, model4]);

      var appliedFilters = eventSearch.getAppliedFilters();
      expect(appliedFilters.length).toEqual(4);
      expect(appliedFilters[0].values).toEqual(['value 0']);
      expect(appliedFilters[1].values).toEqual(['value 1']);
      expect(appliedFilters[2].values).toEqual(['value 2']);
      var values = appliedFilters[3].values;
      expect(values.length).toEqual(2);
      expect(values[0]).toEqual('value 4.1');
      expect(values[1]).toEqual('value 4.2');
    });

    it('can apply filters to the view model', function() {
      spyOn(eventSearch.viewModel, 'appendFilter');
      eventSearch.applyFilter('filter', 'compareType', 'value');
      expect(eventSearch.viewModel.appendFilter).wasCalledWith('filter', 'compareType', 'value');
    });

    it('can clear events', function() {
      eventSearch.events([1, 2, 3]);
      eventSearch.clear();
      expect(eventSearch.events().length).toEqual(0);
    });

    it('can find filters', function() {
      var filter1 = { propertyName: 'filter1' };
      var filter2 = { propertyName: 'filter2' };
      var filter3 = { propertyName: 'filter3' };
      eventSearch.filters([filter1, filter2, filter3]);
      var foundFilter = eventSearch.findFilter('filter2');
      expect(foundFilter).toEqual(filter2);
      foundFilter = eventSearch.findFilter('unknown');
      expect(foundFilter).toEqual(null);
    });

    it('can register filters', function() {
      var filter1 = {};
      expect(eventSearch.filters().length).toEqual(0);
      eventSearch.registerFilter(filter1);
      expect(eventSearch.filters().length).toEqual(1);
      expect(eventSearch.filters()[0]).toEqual(filter1);

      var filter2 = {};
      var filter3 = {};
      eventSearch.registerAllFilters([filter2, filter3]);
      expect(eventSearch.filters().length).toEqual(3);
    });

    it('won\'t register a filter that is in the static filters list', function() {
      var filter1 = {};
      expect(eventSearch.filters().length).toEqual(0);
      eventSearch.registerFilter(filter1);
      expect(eventSearch.filters().length).toEqual(1);
      expect(eventSearch.filters()[0]).toEqual(filter1);
      
      var staticFilter = {propertyName: 'static'};
      eventSearch.staticFilters = [staticFilter];

      var filter2 = {propertyName: 'static'};
      var filter3 = {};
      eventSearch.registerAllFilters([filter2, filter3]);
      expect(eventSearch.filters().length).toEqual(2);
      expect(_.find(eventSearch.filters(), function(filter) {
        return staticFilter.propertyName === filter.propertyName;
      })).toBeUndefined();
    });

    it('adds utility methods to registered filters', function() {
      var filter = {};
      eventSearch.registerFilter(filter);
      expect(filter.has).toBeDefined();
      expect(filter.hasnt).toBeDefined();
      expect(filter.executeQuery).not.toBeDefined();

      // Test the utility methods.
      filter.catpants = true;
      expect(filter.has('catpants')).toBeTruthy();
      expect(filter.has('catpants', 'doghat')).toBeFalsy();
      expect(filter.hasnt('catpants')).toBeFalsy();
      expect(filter.hasnt('catpants', 'doghat')).toBeTruthy();
    });

    describe('executeQuery', function() {
      var filter = {
        valueUri: 'some/made/up/url'
      };
      var mockQuery;

      beforeEach(function() {
        mockQuery = {
          term: '',
          callback: jasmine.createSpy()
        };
        eventSearch.registerFilter(filter);
        expect(filter.executeQuery).toBeDefined();
      });

      it('should not make request for empty term', function() {
        filter.executeQuery(mockQuery);
        var request = mostRecentAjaxRequest();
        expect(request.url).not.toEqual(filter.valueUri);
        expect(mockQuery.callback).wasNotCalled();
      });

      it('should make request for partial term', function() {
        mockQuery.term = 'cat';
        filter.executeQuery(mockQuery);
        request = mostRecentAjaxRequest();
        expect(request.url).toEqual(filter.valueUri + '?q=cat');
        var response = [
          'Sir Catpants',
          'Son of Catpants'
        ];
        request.response({
          status: 200,
          responseText: JSON.stringify(response)
        });
        expect(mockQuery.callback).wasCalled();
        var args = mockQuery.callback.mostRecentCall.args;
        expect(args.length).toEqual(1);
        var results = args[0].results;
        expect(results).toBeDefined();
        expect(results.length).toEqual(2);
        var result = results[0];
        expect(result.id).toEqual(response[0]);
        expect(result.text).toEqual(response[0]);
        result = results[1];
        expect(result.id).toEqual(response[1]);
        expect(result.text).toEqual(response[1]);
      });
    });

    it('can clear out the filters', function() {
      var filter1 = { propertyName: 'filter1' };
      var filter2 = { propertyName: 'filter2' };
      var filter3 = { propertyName: 'filter3' };
      eventSearch.filters([filter1, filter2, filter3]);
      expect(eventSearch.filters().length).toEqual(3);
      eventSearch.clearFilters();
      expect(eventSearch.filters().length).toEqual(0);
    });
  });
});
